static sreg_t error_noperm(struct xen_cpu *cpu, ureg_t *args)
{
    return -EPERM; /* we don't do dom0 */
}

static sreg_t error_nosys(struct xen_cpu *cpu, ureg_t *args)
{
    return -ENOSYS; /* not implemented */
}

static sreg_t stack_switch(struct xen_cpu *cpu, ureg_t *args)
{
    cpu->kernel_ss = fix_sel(args[0]);
    cpu->kernel_sp = args[1];

#if 0 == longmode
    cpu->tss.ss1  = cpu->kernel_ss;
    cpu->tss.esp1 = cpu->kernel_sp;
#endif
    return 0;
}

static sreg_t update_descriptor(struct xen_cpu *cpu, ureg_t *args)
{
#if longmode
    uint64_t pa = args[0];
    struct descriptor_32 desc = {
	.a = args[1] & 0xffffffff,
	.b = args[1] >> 32,
    };
#else
    uint64_t pa   = args[0] | (uint64_t)args[1] << 32;
    struct descriptor_32 desc = {
	.a = args[2],
	.b = args[3],
    };
#endif
    struct descriptor_32 *guest_gdt;
    int p, index;
    uint64_t mfn;

    fix_desc(&desc);

    mfn = addr_to_frame(pa);
    for (p = 0; p < 16; p++) {
	if (mfn == cpu->gdt_mfns[p])
	    break;
    }
    if (p == 16) {
	printk(1, "%s: not found in gdt: pa %" PRIx64 " (ldt update?)\n",
	       __FUNCTION__, pa);
    } else {
	/* update emu gdt shadow */
	index = addr_offset(pa) / sizeof(struct descriptor_32);
	cpu->gdt[p * 512 + index] = desc;
    }
    
    /* update guest gdt/ldt */
    guest_gdt = map_page(pa);
    *guest_gdt = desc;
    free_page(guest_gdt);
    return 0;
}

static sreg_t fpu_taskswitch(struct xen_cpu *cpu, ureg_t *args)
{
    if (args[0]) {
	write_cr0(X86_CR0_TS|read_cr0());
    } else {
	clts();
    }
    return 0;
}

static sreg_t grant_table_op(struct xen_cpu *cpu, ureg_t *args)
{
    struct gnttab_setup_table *st;
    struct gnttab_query_size  *qs;
    unsigned long *frames;
    int i, rc = 0;
    
    switch (args[0]) {
    case GNTTABOP_setup_table:
	st = (void*)args[1];
	printk(1, "%s: setup_table %d\n", __FUNCTION__, st->nr_frames);
	if (st->nr_frames > GRANT_FRAMES_MAX) {
	    st->status = GNTST_general_error;
	} else {
	    grant_frames = st->nr_frames;
	    frames = st->frame_list.p;
	    for (i = 0; i < grant_frames; i++)
		frames[i] = EMU_MFN(grant_table) + i;
	    st->status = GNTST_okay;
	}
	break;
    case GNTTABOP_query_size:
	printk(1, "%s: query_size\n", __FUNCTION__);
	qs = (void*)args[1];
	qs->nr_frames = grant_frames;
	qs->max_nr_frames = GRANT_FRAMES_MAX;
	qs->status = GNTST_okay;
	break;
    default:
	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	rc = -ENOSYS;
    }
    return rc;
}

static sreg_t xen_version(struct xen_cpu *cpu, ureg_t *args)
{
    switch (args[0]) {
    case XENVER_version:
	return (3 << 16) | 1;

    case XENVER_get_features:
    {
	xen_feature_info_t *fi = (void*)args[1];
	fi->submap = 0;
	if (0 == fi->submap_idx)
	    fi->submap |= (1 << XENFEAT_pae_pgdir_above_4gb);
 	break;
    }

#if 0 == longmode
    case XENVER_platform_parameters:
    {
	uint32_t *ptr32 = (void*)args[1];
	*ptr32 = is_pae() ? XEN_M2P_PAE : XEN_M2P_32;
	break;
    }
#endif

    default:
	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t vm_assist(struct xen_cpu *cpu, ureg_t *args)
{
    int type = args[1];

    switch (args[0]) {
    case VMASST_CMD_enable:
	printk(1, "%s: enable %d\n", __FUNCTION__, type);
	break;
    case VMASST_CMD_disable:
	printk(1, "%s: disable %d\n", __FUNCTION__, type);
	break;
    default:
	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t sched_op(struct xen_cpu *cpu, ureg_t *args)
{
    switch (args[0]) {
    case SCHEDOP_yield:
	/* Hmm, some day, on SMP, we want probably do something else ... */
	halt();
	break;

    case SCHEDOP_block:
	halt();
	break;

    case SCHEDOP_shutdown:
	shutdown();
	break;

    default:
	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t memory_op(struct xen_cpu *cpu, ureg_t *args)
{
    int cmd = args[0] & 0x0f /* MEMOP_CMD_MASK */;
    
    switch (cmd) {
    case XENMEM_increase_reservation:
    {
	struct xen_memory_reservation res;

	if (0 != memcpy_pf(&res, (void*)(args[1]), sizeof(res)))
	    return -EFAULT;
	if (res.domid != DOMID_SELF)
	    return -EPERM;
	printk(0, "%s: increase_reservation: nr %ld, order %d (not implemented)\n",
	       __FUNCTION__, res.nr_extents, res.extent_order);
	/* FIXME: not implemented yet, thus say "no pages allocated" */
	return 0;
    }
    case XENMEM_decrease_reservation:
    {
	struct xen_memory_reservation res;
	xen_pfn_t *ptr, gmfn;
	int i, p, count = 0;

	if (0 != memcpy_pf(&res, (void*)(args[1]), sizeof(res)))
	    return -EFAULT;
	if (res.domid != DOMID_SELF)
	    return -EPERM;
	ptr = res.extent_start.p;
	for (i = 0; i < res.nr_extents; i++) {
	    if (0 != memcpy_pf(&gmfn, ptr + i, sizeof(gmfn)))
		break;
	    for (p = 0; p < (1 << res.extent_order); p++)
		m2p[gmfn+p] = INVALID_M2P_ENTRY;
	    /* FIXME: make host free pages */
	    count++;
	}
	printk(2, "%s: decrease_reservation: nr %ld, order %d"
	       " (max %" PRIx64 "/%" PRIx64 ") -> rc %d\n", __FUNCTION__,
	       res.nr_extents, res.extent_order,
	       vminfo.pg_guest, vminfo.pg_total,
	       count);
	vminfo.balloon += count;
	return count;
    }
    case XENMEM_populate_physmap:
    {
	struct xen_memory_reservation res;
	xen_pfn_t *ptr, gpfn, gmfn;
	int i, p, count = 0;

	if (0 != memcpy_pf(&res, (void*)(args[1]), sizeof(res)))
	    return -EFAULT;
	if (res.domid != DOMID_SELF)
	    return -EPERM;
	ptr = res.extent_start.p;
	for (i = 0; i < res.nr_extents; i++) {
	    if (0 != memcpy_pf(&gpfn, ptr + i, sizeof(gpfn)))
		break;
	    gmfn = gpfn + vminfo.mfn_guest;
	    for (p = 0; p < (1 << res.extent_order); p++) {
		if (m2p[gmfn+p] != INVALID_M2P_ENTRY) {
		    /* trying to balloon up might trigger this */
		    printk(1, "%s:    %4d: gpfn %08lx, gmfn %08lx: %s\n", __FUNCTION__,
			   i, gpfn+p, gmfn+p, "Huh? Page isn't free!");
		    break;
		}
		m2p[gmfn+p] = gpfn+p;
	    }
	    if (p != (1 << res.extent_order))
		break;
	    /* FIXME: make host reclaim pages */
	    if (0 != memcpy_pf(ptr + i, &gmfn, sizeof(gmfn)))
		break;
	    count++;
	}
	printk(2, "%s: populate_physmap: nr %ld, order %d -> rc %d\n",
	       __FUNCTION__, res.nr_extents, res.extent_order, count);
	vminfo.balloon += count;
	return count;
    }
    case XENMEM_machphys_mapping:
    {
	struct xen_machphys_mapping map;
	uint32_t pg_m2p = 1024 /* pages (4 MB) */;
	void *dest = (void*)(args[1]);
#if longmode
	map.v_start = XEN_M2P_64;
#else
	if (is_pae()) {
	    map.v_start = XEN_M2P_PAE;
	} else {
	    map.v_start = XEN_M2P_32;
	}
#endif
	map.v_end   = map.v_start + frame_to_addr(pg_m2p);
	map.max_mfn = pg_m2p << (PAGE_SHIFT-3);
	if (0 != memcpy_pf(dest, &map, sizeof(map)))
	    return -EFAULT;
	return 0;
    }
    
    case XENMEM_memory_map:
	/* we have no e820 map */
	return -ENOSYS;

    default:
	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, cmd);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t set_trap_table(struct xen_cpu *cpu, ureg_t *args)
{
    struct trap_info *traps;
    struct trap_info trap;
    int i;

    if (0 == args[0]) {
	printk(0, "%s: FIXME: clear table\n", __FUNCTION__);
	return -EINVAL;
    }

    traps = (void*)args[0];
    for (i = 0;; i++) {
	if (0 != memcpy_pf(&trap, traps+i, sizeof(trap)))
	    return -EFAULT;
	if (!trap.address)
	    break;
	trap.cs = fix_sel32(trap.cs);
	xentr[traps[i].vector] = trap;
#if 0 == longmode
	if (traps[i].vector > 0x40) {
	    /* route directly */
	    uint32_t dpl = trap.flags & 0x03;
	    xen_idt[traps[i].vector] =
		mkgate32(trap.cs, trap.address, 0x8f | (dpl << 5));
	}
#endif
    }
    return 0;
}

static void callback_setup(int type, xen_callback_t *cb)
{
#if 0 == longmode
    cb->cs = fix_sel32(cb->cs);
#endif
    xencb[type] = *cb;
}

static void callback_clear(int type)
{
#if longmode
    xencb[type] = 0;
#else
    xencb[type].cs = 0;
    xencb[type].eip = 0;
#endif
}

static sreg_t set_callbacks(struct xen_cpu *cpu, ureg_t *args)
{
#if longmode
    callback_setup(CALLBACKTYPE_event, &args[0]);
    callback_setup(CALLBACKTYPE_failsafe, &args[1]);
#else
    xen_callback_t cb;

    cb.cs  = args[0];
    cb.eip = args[1];
    callback_setup(CALLBACKTYPE_event, &cb);
    cb.cs  = args[2];
    cb.eip = args[3];
    callback_setup(CALLBACKTYPE_failsafe, &cb);
#endif
    return 0;
}

static sreg_t callback_op(struct xen_cpu *cpu, ureg_t *args)
{
    struct callback_register cb;

    memcpy_pf(&cb, (void*)(args[1]), sizeof(cb));
    if (cb.type >= 8)
	return -EINVAL;
    
    switch (args[0]) {
    case CALLBACKOP_register:
	callback_setup(cb.type, &cb.address);
	if (cb.flags)
	    printk(0, "%s: FIXME: ignored flags\n", __FUNCTION__);
	break;
    case CALLBACKOP_unregister:
	callback_clear(cb.type);
	break;
    default:
	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

int guest_gdt_init(struct xen_cpu *cpu, uint32_t entries, ureg_t *mfns)
{
    uint32_t pages = (entries + 511) / 512;
    struct descriptor_32 *src, *dst, tmp;
    int p,e;

    for (p = 0; p < pages; p++) {
	cpu->gdt_mfns[p] = mfns[p];
	src = map_page(frame_to_addr(mfns[p]));
	dst = cpu->gdt + p * 512;
	for (e = 0; e < 512; e++) {
	    tmp = src[e];
	    fix_desc(&tmp);
	    dst[e] = tmp;
	}
	free_page(src);
    }
    return 0;
}

static sreg_t set_gdt(struct xen_cpu *cpu, ureg_t *args)
{
    ureg_t mfns[16];
    uint32_t entries = args[1];

    if (0 != memcpy_pf(mfns, (void*)(args[0]), sizeof(mfns)))
	return -EFAULT;
    if (entries > (0xe000 >> 3))
	return -EINVAL;

    return guest_gdt_init(cpu, entries, mfns);
}

static sreg_t vcpu_op(struct xen_cpu *cpu, ureg_t *args)
{
    struct xen_cpu *vcpu;
    
    switch (args[0]) {
    case VCPUOP_initialise:
	printk(0, "%s: initialise cpu %d\n", __FUNCTION__, (int)args[1]);
	vcpu = cpu_find(args[1]);
	if (!vcpu->init_ctxt)
	    vcpu->init_ctxt = get_memory(sizeof(*(vcpu->init_ctxt)), "init_ctxt");
	if (0 != memcpy_pf(vcpu->init_ctxt, (void*)args[2],
			   sizeof(*(vcpu->init_ctxt))))
	    return -EFAULT;
	return 0;
    default:
	/* forward to userspace */
	return -200;
    }
}
